/*  FrvRes.c by  B.Mours Sept 29, 2000           */


#include "FrvRes.h"
#include <string.h>
#include <math.h>

/*-------------------------------------------------------------- FrvResFree -*/
void FrvResFree(FrvRes *res)
/*---------------------------------------------------------------------------*/
{
 if(res == NULL) return;

 free(res->name);

 FrVectFree(res->sum);
 FrVectFree(res->sum2);
 FrVectFree(res->mean);
 FrVectFree(res->residual);
 FrVectFree(res->oneOverRms);


 return;
}
/*--------------------------------------------------------------- FrvResIni -*/
int FrvResIni(FrvRes *res, FrVect *vect)
/*---------------------------------------------------------------------------*/
/*   Returns: 0 in case of probleme, 1 otherwise                             */
/*---------------------------------------------------------------------------*/
{int i;

 if(res == NULL) return(0);
 if(vect == NULL) return(0);
       
                /*--- save vector name -----*/

 if(FrStrCpy(&(res->name),vect->name) == NULL) return(0);

               /*--------- create output structures -----*/

 res->sum        = FrvCopyToD(vect,"sum");
 res->sum2       = FrvCopyToD(vect,"sum2");
 res->mean       = FrvCopyToD(vect,"mean");
 res->residual   = FrvCopyToD(vect,"residual");
 res->oneOverRms = FrvCopyToD(vect,"oneOverRms");

 if((res->sum        == NULL) ||
    (res->sum2       == NULL) ||
    (res->mean       == NULL) ||
    (res->residual   == NULL) ||
    (res->oneOverRms == NULL))   return(0);
    
 for (i=0; i<res->sum->nData; i++)  
   {res->sum->dataD[i]       = 0.;
    res->sum2->dataD[i]      = 0.;
    res->oneOverRms->dataD[i] = 1.;}

 return(1);
}
/*--------------------------------------------------------------- FrvResNew -*/
FrvRes* FrvResNew(int period, double scale)
/*---------------------------------------------------------------------------*/
{FrvRes *res;

 res = (FrvRes *) (calloc(1, sizeof(FrvRes)));
 if(res == NULL) return(NULL);

 res->period = period;
 res->scale  = scale;

 res->nCall = 0;
                        
 return(res);
}
/*--------------------------------------------------------------- FrvResProc-*/
int FrvResProc(FrvRes *res, FrVect *vect)
/*---------------------------------------------------------------------------*/
{int i;
 double *sum, *sum2, *mean, *oneOverRms, *residual,
        value, weigth, rms2;

 if(res  == NULL)  return(0);
 if(vect == NULL)  return(0);

          /*------------ initialized res object if needed ------------------*/

 if(res->nCall == 0)
   {if(FrvResIni(res,vect) == 0) return(0);}

 res->nCall++;

          /*---------- check that we work with the same type of vector -----*/

 if(vect->nData != res->sum->nData) return(0);

         /*--------------- rebuilt the means and res values if needed------*/

 sum        = res->sum->dataD;
 sum2       = res->sum2->dataD;
 mean       = res->mean->dataD;
 oneOverRms = res->oneOverRms->dataD;
 residual   = res->residual->dataD;

 if(((res->nCall)%(res->period)) == 0)
   {weigth = 1./res->period;
    for(i=0; i<vect->nData; i++)
      {mean[i] = sum[i] * weigth;
      rms2     = sum2[i]* weigth - mean[i]*mean[i];
      if(rms2 > 0.)
	    {oneOverRms[i] = res->scale/sqrt(rms2);}
       else {oneOverRms[i] = 1.;}
       sum[i] = 0.;
       sum2[i] = 0.;}

    res->mean->GTimeS = vect->GTimeS;
    res->mean->GTimeN = vect->GTimeN;
    res->oneOverRms->GTimeS = vect->GTimeS;
    res->oneOverRms->GTimeN = vect->GTimeN;}

          /*------------------------------- now loop over the samples ------*/

 for(i=0; i<vect->nData; i++)
   {if     (vect->type == FR_VECT_2S) {value = vect->dataS[i];}
    else if(vect->type == FR_VECT_8R) {value = vect->dataD[i];}
    else if(vect->type == FR_VECT_4R) {value = vect->dataF[i];}
    else if(vect->type == FR_VECT_C)  {value = vect->data[i];}
    else if(vect->type == FR_VECT_4S) {value = vect->dataI[i];}
    else if(vect->type == FR_VECT_8S) {value = vect->dataL[i];}
    else if(vect->type == FR_VECT_2U) {value = vect->dataUS[i];}
    else if(vect->type == FR_VECT_4U) {value = vect->dataUI[i];}
    else if(vect->type == FR_VECT_8U) {value = vect->dataUL[i];}
    else if(vect->type == FR_VECT_1U) {value = vect->dataU[i];}
    else {return(0);}
    
    sum[i]  += value;
    sum2[i] += value*value;

    residual[i] = (value - mean[i])*oneOverRms[i];}

                     /*------- Set start time for all the series defined --- */

 res->residual->GTimeS = vect->GTimeS;
 res->residual->GTimeN = vect->GTimeN;

 return(1);

}


