#include <tid.h>
#include <framecpp/FrAdcData.hh>
#include <Dacc.hh>
#include <DaccIn.hh>
#include <math.h>
#include <errors.h>
#include <TidString.h>

#define _TIDG_

//#define _NODACC_

Tid::Tid () { // default constructor

  DMTBackGround = 0;

  In = NULL;

  TidCon();

}
  
Tid::Tid (Dacc* din) {

  DMTBackGround = 1;

  In = din;

  In->synch();

  TidCon ();
}

void Tid::TidCon () // constructor
  // look for ./.tid, then for $HOME/.tid. If none, set default values
{
  maxClusterSize = 30; // default

  ContinueLoop = 1;
  T0Set = 0;
  IO = NULL;
  broadcaster=NULL;
  IOin = NULL;

  // allocate pool of events
  static const unsigned int EPoolSize = 1024;
  for(unsigned int i = 0; i < EPoolSize; i++) {
    Event* tev = new Event ();
    ePool.push(tev);
  }

  if(tidg != NULL) 
    throw TidError ("Trying to create a Tid object while one already exists.");

  tidg = this;

  FILE *in;

  // try this directory
  bool pSet = 1;
  char *pwd;

  try {

  if((pwd = getenv("PWD")) != NULL) {
    char buf[2048];
    sprintf(buf,"%s/.tid",pwd);
    if((in = fopen(buf,"r")) != NULL) {
      try {
	SetParamsFromFile(in);
      }
      catch (TidError err) {HandleError(err);}
      fclose(in);
      pSet = 0;
    }
  }

  // try home
  char* home;
  if(pSet) {
    if((home = getenv("HOME")) != NULL) {
      char buf[2048];
      sprintf(buf,"%s/.tid",home);
      if((in = fopen(buf,"r")) != NULL) {
	try {
	  SetParamsFromFile(in);
	}
	catch (TidError err) {HandleError(err);}
	fclose(in);
	pSet = 0;
      }
    }
  }

  // set default parameters
  if(pSet) {
    sName = string("No Server");

    TidMessage(string("Connecting to data source (") +
	       sName +
	       string(")."));
  
    In = new Dacc(sName.c_str());
    if(!In) throw TidError("FATAL:Can't connect to data server.");

    IO = new TidIO(2077); // launch data server
    broadcaster = new TidBroadcast(*IO);

    IOin = new TidIO(2078); // launch message server

    TidMessage("Setting default parameter values.");
    
    tres = 1.0;

    // writes default file in $HOME, or in working directory if $HOME failed
    if(home != NULL) {
      char buf[2048];
      sprintf(buf,"%s/.tid",home);
      if((in = fopen(buf,"w")) != NULL) {
	WriteParamsToFile(in);
	fclose(in);
      }
    }
    else {
      if((in = fopen("./.tid","w")) != NULL) {
	WriteParamsToFile(in);
	fclose(in);
      }
    }
  }

  /* this is now in fileio.cc
  TidMessage(string("Connecting to data source (") +
	     sName +
	     string(")."));
  
  In = new Dacc(sName.c_str());
  */

  list<ChnInfo*>::iterator p;
  for(p=diag.begin(); p != diag.end(); p++) {
    TSeries** ttmp = new TSeries* (0x0);
    data.push_back(ttmp);


    unsigned int trueFs = GetSamplingByName((*p)->GetName());
    unsigned int decim;

    if(trueFs == (*p)->GetfHigh()) decim = 0;
    else {
      decim = (unsigned int)pow(2.0, floor(log(0.5 * (double)trueFs / (*p)->GetfHigh())/log(2.0)));
      (*p)->SetFs(trueFs / decim);
    }

    In->addChannel((*p)->GetName().c_str(),
		   decim,
		   ttmp);

    char dfa[64];
    sprintf(dfa,"%u",decim);

    TidMessage ( string("Adding channel ") + (*p)->GetName() +
		 string(" with decimation factor ") + dfa );
  }
   

  /*
    In->addChannel((*p)->GetName().c_str(),
		   0,
		   ttmp);
    TidMessage ( string("Adding channel ") + (*p)->GetName() +
		   string(" with no decimation. "));
  }
  */
  }
  catch (TidError err) {HandleError(err);}

  TidMessage("Computation server is up.");
}


Tid::~Tid () // destructor
{
  //  TidMessage("Computation server is shutting down.");

  for(unsigned int i=0; i<ePool.size(); i++) {
    delete ePool.top();
    ePool.pop();
  }

  if(In) delete In;
  if(broadcaster) delete broadcaster;
  if(IO) delete IO;
  if(IOin) delete IOin;

  cerr<<"Computation server is down.\n";
}


unsigned int Tid::GetSamplingByName(const string &chnName) { // return Nyquist frequency of channel named chnName

  unsigned int trueFs;
  
#ifndef _NODACC_  
  const FrameCPP::FrAdcData* adc = In->findAdc(chnName);
  if(adc == NULL)
    throw TidError ( string("Channel ") + chnName +
		  string(" not found."));
  else {
    trueFs = (unsigned int)adc->getSampleRate();  
  }
  
#else  
  trueFs = 256;
#endif

  return trueFs;

}




bool Tid::Loop () // main computation loop
{
  // check for requests for new connection
  int nsock;

  try {

    if((nsock = broadcaster->Listen()) >= 0) 
      broadcaster->SendTableofContent(this, nsock);

    if((nsock = IOin->Listen()) >= 0) {
      if((IOin->GetConnected() == 1)) {
	TidString welcome("Tid server running at ");
	welcome += sName;
	welcome += "\n";
	IOin->WritePlain(welcome);
      }
      else {
	TidString toomany("Tid server running at ");
	toomany += sName;
	toomany += "\nA user is already connected. Aborting.";
	IOin->WritePlain(toomany);
	IOin->HangUp(nsock);
      }
    }

    TidString* tmpstr;
    while(IOin->ReadPlain(&tmpstr)) {
      ProcessInstructions(*tmpstr);
      delete tmpstr;
    }
  }
  catch (TidError err) {HandleError(err);}


  static int t = 0;

  cout<<t<<" "<<flush;

  try {

    bool OSCPass = 1;
    list<string>::iterator osc;

    for(osc=oscConditions.begin(); osc!=oscConditions.end(); osc++)
      if(!osclist->satisfied((*osc).c_str())) {OSCPass=0; break;}
    
    int rc(0); bool newfill(true);
    do {rc=In->fillData(tres, newfill); newfill=false;} while (rc==-8); 
    switch(rc) {
    case 0:
      break;
    case -1:
      throw TidError ("FATAL:Frame start not contiguous to previous data.");
      break;
    case -2:
      throw TidError ("FATAL:Sample rate isn't compatible with previous data.");
      break;
    case -3:
      throw TidError ("FATAL:Requested data not found in current frame.");
      break;
    case -4:
      throw TidError ("FATAL:Error reading frame.");
      break;
    case -5:
      throw TidError ("FATAL:Frame data are not self-consistent.");
      break;
    case -6:
      throw TidError ("FATAL:TSeries is not allocated.");
      break;
    case -7:
      throw TidError ("FATAL:Unsupported data type.");
      break;
    default:
      throw TidError ("FATAL:Unknown value returned by fill_Data.");
      break;
    }

    list<ChnInfo*>::iterator p;
    list<TSeries**>::iterator q = data.begin();
    list<unsigned int>::iterator i;

    if(!T0Set) {
      T0Set = 1;
      T0 = (*(*q))->getStartTime();
    }
   
    for(p = diag.begin(), i=cidlist.begin(); 
	p != diag.end(); 
	p++, q++, i++) {

      if((*p)->DoFilter()) {
	**(*q) = (*p)->GetFilter()->apply(**(*q));
      }

      if(OSCPass || !((*p)->DoOSC()))
	{
	  **(*q) -= (*(*q))->getAverage(); // remove mean
	  FSeries tmpD = FSeries(**(*q));

	  try {
	    (*p)->ProcessData(*i, &tmpD);
	  }
	  catch (TidError err) {HandleError(err);}

	}
      else if(!OSCPass && (*p)->DoOSC()) {
	//	TidWarning(string("Lost lock. Recomputing threshold for channel ") + (*p)->GetName());
	(*p)->ComputeThreshold();
      }
    }
  }
    
  catch (TidError err) {HandleError(err);}
  
  t++;

  return ContinueLoop;
}


void Tid::DeleteEvent (Event *ev) { // clean Event and put pointer back into pool

  ev->Reset();

  ePool.push(ev);

}


Event* Tid::NewEvent () { // return pointer to Event

  if(ePool.empty()) {
    // allocate pool of events
    static const unsigned int EPoolSize = 1024;
    for(unsigned int i = 0; i < EPoolSize; i++) {
      Event* tev = new Event ();
      ePool.push(tev);
    }
  }

  Event* ret = ePool.top();

  ePool.pop();

  return ret;

}


void catch_sig(int sig)
{
  switch (sig)
    {
    case SIGHUP:
      TidMessage("Hangup! ");
      break;
    case SIGINT:
      TidMessage("Interrupt! ");
      break;
    case SIGQUIT:
      TidMessage("Quit! ");
      break;
    case SIGABRT:
      TidMessage("Abort! ");
      break;
    case SIGTERM:
      TidMessage("Terminate process!");
      break;
    default:
      TidMessage("Unknown signal! ");
  }
  TidMessage("Shutting down because of signal");

  if(tidg) delete tidg;
  tidg = NULL;

  exit(-1);
}


void Tid::AddMessage (const string& s) {

  messages.push(s);

  if(IOin)
    IOin->WritePlain(s);

}


void Tid::AddComplain(const string& s) {

  complains.push(s);

  if(IOin)
    IOin->WritePlain(s);

}


void Tid::ProcessInstructions(TidString &A) {

  if(A == "exit" ||
     A == "quit") {
    IOin->WritePlain(string("Closing connection to tid server."));
    IOin->HangUp();
    return;
  }

  const char* buf = A.c_str();

  if(strstr(buf,"delete") == buf) {
    char* cname;
    if((cname = strchr(buf,32))==NULL) {
      TidWarning ("Malformed command: delete <channelname>");
      return;
    }
    cname++;
    string B(cname);

    list<ChnInfo*>::iterator p;
    list<TSeries**>::iterator q;
    list<unsigned int>::iterator i;

    bool knownchn = 0;

    for(p=diag.begin(), q=data.begin(), i=cidlist.begin(); 
	p!=diag.end(); 
	p++, q++, i++) {
      if((*p)->GetName() == B) {
	knownchn = 1;

	In->rmChannel(B.c_str());

	delete *q;
	
	p=diag.erase(p);
	q=data.erase(q);
	i=cidlist.erase(i);

	if(p!=diag.begin()) {p--; q--; i--;}

	TidMessage(string("Channel ") + B + string(" deleted."));
      }
    }

    if(!knownchn)
      TidWarning(string("Unknown channel: ") + B);

    return;
  }




  if(strstr(buf,"reset") == buf) {
    char* cname;
    if((cname = strchr(buf,32))==NULL) {
      TidWarning ("Malformed command: reset <channelname>");
      return;
    }
    cname++;
    string B(cname);

    list<ChnInfo*>::iterator p;

    bool knownchn = 0;

    for(p=diag.begin();
	p!=diag.end(); 
	p++) {
      if((*p)->GetName() == B) {
	knownchn = 1;
	TidWarning(string("Forcing recomputation of thresholds for channel ") + B);
	(*p)->ComputeThreshold();

      }
    }

    if(!knownchn)
      TidWarning(string("Unknown channel: ") + B);

    return;
  }




  if(strstr(buf,"scale") == buf) {
    char* cname;
    char* sfb;

    if((cname = strchr(buf,32))==NULL ||
       (sfb = strchr(cname+1,32))==NULL) {
      TidWarning ("Malformed command: scale <channelname> <scaleFactor>");
      return;
    }
    cname++;
    *sfb = 0;
    sfb++;
    string B(cname);
    float sf = atof(sfb);

    list<ChnInfo*>::iterator p;

    bool knownchn = 0;

    for(p=diag.begin();
	p!=diag.end(); 
	p++) {
      if((*p)->GetName() == B) {
	knownchn = 1;
	TidWarning(string("Scaling threshold for channel ") + B);
	(*p)->ScaleThreshold(sf);

      }
    }

    if(!knownchn)
      TidWarning(string("Unknown channel: ") + B);

    return;
  }




  if(strstr(buf,"bpp") == buf) {
    char* cname;
    char* sfb;

    if((cname = strchr(buf,32))==NULL ||
       (sfb = strchr(cname+1,32))==NULL) {
      TidWarning ("Malformed command: bpp <channelname> <BPP>");
      return;
    }
    cname++;
    *sfb = 0;
    sfb++;
    string B(cname);
    float sf = atof(sfb);

    list<ChnInfo*>::iterator p;

    bool knownchn = 0;

    for(p=diag.begin();
	p!=diag.end(); 
	p++) {
      if((*p)->GetName() == B) {
	knownchn = 1;
	TidWarning(string("Changing BPP for channel ") + B);
	(*p)->SetBPP(sf);

      }
    }

    if(!knownchn)
      TidWarning(string("Unknown channel: ") + B);

    return;
  }



  if(strstr(buf,"add") == buf) {
    char *pos, *pos1;
    char buff[1024];

    if((pos1 = strchr(buf,32))==NULL) {
      TidWarning ("Malformed command: add <channelname> <LowFreq> <HighFreq> <StabTimeScale> <StabTol> <StabTestFreq> <Sigma> <Delta> <Alpha> <BPP>");
      return;
    }
    pos1++;
    

#define GETBUF     pos = pos1; \
                   if((pos1 = strchr(pos+1,32))==NULL) { \
		     TidWarning ("Malformed command: add <channelname> <LowFreq> <HighFreq> <StabTimeScale> <StabTol> <StabTestFreq> <Sigma> <Delta> <Alpha> <BPP>"); \
                     return; \
                   } \
                   pos1++; \
                   memcpy(buff,pos,pos1-pos); \
                   buff[pos1-pos-1]='\x0'; 

    GETBUF;					  
    string chnname(buff);
    unsigned int fs = GetSamplingByName(chnname);

    GETBUF;
    double lofreq = atof(buff);

    GETBUF;
    double hifreq = atof(buff);
    unsigned int dec = (unsigned int)ceil((double)fs/hifreq);

    GETBUF;
    double stabscale = atof(buff);

    GETBUF;
    double stabtol = atof(buff);

    GETBUF;
    vector<unsigned int> testfreq;
    if(strcmp(buff,"All")==0) {
      for(unsigned int i=0; i<fs/2; i++)
	testfreq.push_back(i);
    }
    else {
      char *ptr, *ptr2;
      ptr = buff;
      ptr2 = strchr(ptr,',');
      double BW = 1.0/tres; 

      while(ptr2 != NULL) {
	*ptr2 = '\x0';
	testfreq.push_back((unsigned int)floor(atoi(ptr)/BW));

	ptr = ptr2+1;
	ptr2 = strchr(ptr,',');
      }

      testfreq.push_back((unsigned int)floor(atoi(ptr)/BW));
    }

    GETBUF;
    unsigned sigma = (unsigned int)atoi(buff);

    GETBUF;
    vector<unsigned int> delta;
    char *ptr, *ptr2;
    ptr = buff;
    ptr2 = strchr(ptr,',');
 
    while(ptr2 != NULL) {
      *ptr2 = '\x0';
      delta.push_back((unsigned int)atoi(ptr));
      ptr = ptr2+1;
      ptr2 = strchr(ptr,',');
    }
    delta.push_back((unsigned int)atoi(ptr));

    GETBUF;
    double alpha = atof(buff);

    double bpp = atof(pos1);
 
    AddChannel ( chnname,
		 sigma,
		 delta,
		 alpha,
		 bpp,
		 fs,
		 lofreq,
		 hifreq,
		 dec,
		 tres,
		 stabscale,
		 stabtol,
		 testfreq
		 );

    TSeries** ttmp = new TSeries* (0x0);
    data.push_back(ttmp);

    list<ChnInfo*>::iterator p = diag.end();
    p--;

    unsigned int trueFs = fs;
    unsigned int decim;

    if((*p)->GetFs() <= 0) decim = 0;
    else decim = trueFs / (*p)->GetFs();

    In->addChannel((*p)->GetName().c_str(),
		   decim,
		   ttmp);

    char dfa[64];
    sprintf(dfa,"%u",decim);

    TidMessage ( string("Adding channel ") + (*p)->GetName() +
		 string(" with decimation factor ") + dfa );

    return;
  }

  if(A != string("\n") && !(A.empty()))
    TidWarning (string("Unable to process instruction: ") + A.Gstring());
}



/************************************************************************/





